@@ -1,5 +1,6 @@
 # -*- test-case-name: automat._test.test_methodical -*-
 
+import collections
 from functools import wraps
 from itertools import count
 
@@ -11,10 +12,60 @@ except ImportError:
     from inspect import getargspec as getArgsSpec
 
 import attr
+import six
 
 from ._core import Transitioner, Automaton
 from ._introspection import preserveName
 
+
+ArgSpec = collections.namedtuple('ArgSpec', ['args', 'varargs', 'varkw',
+                                             'defaults', 'kwonlyargs',
+                                             'kwonlydefaults', 'annotations'])
+
+
+def _getArgSpec(func):
+    """
+    Normalize inspect.ArgSpec across python versions
+    and convert mutable attributes to immutable types.
+
+    :param Callable func: A function.
+    :return: The function's ArgSpec.
+    :rtype: ArgSpec
+    """
+    spec = getArgsSpec(func)
+    return ArgSpec(
+        args=tuple(spec.args),
+        varargs=spec.varargs,
+        varkw=spec.varkw if six.PY3 else spec.keywords,
+        defaults=spec.defaults if spec.defaults else (),
+        kwonlyargs=tuple(spec.kwonlyargs) if six.PY3 else (),
+        kwonlydefaults=(
+            tuple(spec.kwonlydefaults.items())
+            if spec.kwonlydefaults else ()
+        ) if six.PY3 else (),
+        annotations=tuple(spec.annotations.items()) if six.PY3 else (),
+    )
+
+
+def _getArgNames(spec):
+    """
+    Get the name of all arguments defined in a function signature.
+
+    The name of * and ** arguments is normalized to "*args" and "**kwargs".
+
+    :param ArgSpec spec: A function to interrogate for a signature.
+    :return: The set of all argument names in `func`s signature.
+    :rtype: Set[str]
+    """
+    return set(
+        spec.args
+        + spec.kwonlyargs
+        + (('*args',) if spec.varargs else ())
+        + (('**kwargs',) if spec.varkw else ())
+        + spec.annotations
+    )
+
+
 def _keywords_only(f):
     """
     Decorate a function so all its arguments must be passed by keyword.
@@ -42,22 +93,35 @@ class MethodicalState(object):
 
     def upon(self, input, enter, outputs, collector=list):
         """
-        Declare a state transition within the L{MethodicalMachine} associated
-        with this L{MethodicalState}: upon the receipt of the input C{input},
-        enter the state C{enter}, emitting each output in C{outputs}.
+        Declare a state transition within the :class:`automat.MethodicalMachine`
+        associated with this :class:`automat.MethodicalState`:
+        upon the receipt of the `input`, enter the `state`,
+        emitting each output in `outputs`.
+
+        :param MethodicalInput input: The input triggering a state transition.
+        :param MethodicalState enter: The resulting state.
+        :param Iterable[MethodicalOutput] outputs: The outputs to be triggered
+            as a result of the declared state transition.
+        :param Callable collector: The function to be used when collecting
+            output return values.
+
+        :raises TypeError: if any of the `outputs` signatures do not match
+            the `inputs` signature.
+        :raises ValueError: if the state transition from `self` via `input`
+            has already been defined.
         """
-        inputSpec = getArgsSpec(input.method)
+        inputArgs = _getArgNames(input.argSpec)
         for output in outputs:
-            outputSpec = getArgsSpec(output.method)
-            if inputSpec != outputSpec:
+            outputArgs = _getArgNames(output.argSpec)
+            if not outputArgs.issubset(inputArgs):
                 raise TypeError(
                     "method {input} signature {inputSignature} "
                     "does not match output {output} "
                     "signature {outputSignature}".format(
                         input=input.method.__name__,
                         output=output.method.__name__,
-                        inputSignature=inputSpec,
-                        outputSignature=outputSpec,
+                        inputSignature=getArgsSpec(input.method),
+                        outputSignature=getArgsSpec(output.method),
                 ))
         self.machine._oneTransition(self, input, enter, outputs, collector)
 
@@ -102,7 +166,47 @@ def assertNoCode(inst, attribute, f):
         raise ValueError("function body must be empty")
 
 
-@attr.s(cmp=False, hash=False)
+def _filterArgs(args, kwargs, inputSpec, outputSpec):
+    """
+    Filter out arguments that were passed to input that output won't accept.
+
+    :param tuple args: The *args that input received.
+    :param dict kwargs: The **kwargs that input received.
+    :param ArgSpec inputSpec: The input's arg spec.
+    :param ArgSpec outputSpec: The output's arg spec.
+    :return: The args and kwargs that output will accept.
+    :rtype: Tuple[tuple, dict]
+    """
+    named_args = tuple(zip(inputSpec.args[1:], args))
+    if outputSpec.varargs:
+        # Only return all args if the output accepts *args.
+        return_args = args
+    else:
+        # Filter out arguments that don't appear
+        # in the output's method signature.
+        return_args = [v for n, v in named_args if n in outputSpec.args]
+
+    # Get any of input's default arguments that were not passed.
+    passed_arg_names = tuple(kwargs)
+    for name, value in named_args:
+        passed_arg_names += (name, value)
+    defaults = zip(inputSpec.args[::-1], inputSpec.defaults[::-1])
+    full_kwargs = {n: v for n, v in defaults if n not in passed_arg_names}
+    full_kwargs.update(kwargs)
+
+    if outputSpec.varkw:
+        # Only pass all kwargs if the output method accepts **kwargs.
+        return_kwargs = full_kwargs
+    else:
+        # Filter out names that the output method does not accept.
+        all_accepted_names = outputSpec.args[1:] + outputSpec.kwonlyargs
+        return_kwargs = {n: v for n, v in full_kwargs.items()
+                         if n in all_accepted_names}
+
+    return return_args, return_kwargs
+
+
+@attr.s(eq=False, hash=False)
 class MethodicalInput(object):
     """
     An input for a L{MethodicalMachine}.
@@ -111,7 +215,11 @@ class MethodicalInput(object):
     method = attr.ib(validator=assertNoCode)
     symbol = attr.ib(repr=False)
     collectors = attr.ib(default=attr.Factory(dict), repr=False)
+    argSpec = attr.ib(init=False, repr=False)
 
+    @argSpec.default
+    def _buildArgSpec(self):
+        return _getArgSpec(self.method)
 
     def __get__(self, oself, type=None):
         """
@@ -132,7 +240,8 @@ class MethodicalInput(object):
             for output in outputs:
                 if outTracer:
                     outTracer(output._name())
-                value = output(oself, *args, **kwargs)
+                a, k = _filterArgs(args, kwargs, self.argSpec, output.argSpec)
+                value = output(oself, *a, **k)
                 values.append(value)
             return collector(values)
         return doInput
@@ -148,6 +257,11 @@ class MethodicalOutput(object):
     """
     machine = attr.ib(repr=False)
     method = attr.ib()
+    argSpec = attr.ib(init=False, repr=False)
+
+    @argSpec.default
+    def _buildArgSpec(self):
+        return _getArgSpec(self.method)
 
     def __get__(self, oself, type=None):
         """
@@ -171,7 +285,7 @@ class MethodicalOutput(object):
     def _name(self):
         return self.method.__name__
 
-@attr.s(cmp=False, hash=False)
+@attr.s(eq=False, hash=False)
 class MethodicalTracer(object):
     automaton = attr.ib(repr=False)
     symbol = attr.ib(repr=False)
@@ -197,8 +311,8 @@ def gensym():
 
 class MethodicalMachine(object):
     """
-    A L{MethodicalMachine} is an interface to an L{Automaton} that uses methods
-    on a class.
+    A :class:`MethodicalMachine` is an interface to an `Automaton`
+    that uses methods on a class.
     """
 
     def __init__(self):
@@ -228,20 +342,18 @@ class MethodicalMachine(object):
         This is a decorator for methods, but it will modify the method so as
         not to be callable any more.
 
-        @param initial: is this state the initial state?  Only one state on
-            this L{MethodicalMachine} may be an initial state; more than one is
-            an error.
-        @type initial: L{bool}
+        :param bool initial: is this state the initial state?
+            Only one state on this :class:`automat.MethodicalMachine`
+            may be an initial state; more than one is an error.
 
-        @param terminal: Is this state a terminal state, i.e. a state that the
-            machine can end up in?  (This is purely informational at this
-            point.)
-        @type terminal: L{bool}
+        :param bool terminal: Is this state a terminal state?
+            i.e. a state that the machine can end up in?
+            (This is purely informational at this point.)
 
-        @param serialized: a serializable value to be used to represent this
-            state to external systems.  This value should be hashable;
-            L{unicode} is a good type to use.
-        @type serialized: a hashable (comparable) value
+        :param Hashable serialized: a serializable value
+            to be used to represent this state to external systems.
+            This value should be hashable;
+            :py:func:`unicode` is a good type to use.
         """
         def decorator(stateMethod):
             state = MethodicalState(machine=self,
@@ -275,7 +387,7 @@ class MethodicalMachine(object):
         This is a decorator for methods.
 
         This method will be called when the state machine transitions to this
-        state as specified in the L{MethodicalMachine.output} method.
+        state as specified in the decorated `output` method.
         """
         def decorator(outputMethod):
             return MethodicalOutput(machine=self, method=outputMethod)
